/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "sc.h"
#include "system.h"
#include "system-common.h"
#include "data.h"
#include "misc.h"
#include "crypto.h"
#include "parse.h"

#define SYSTEM_VIACCESS      0x0500

#define SYSTEM_NAME          "Viaccess"
#define SYSTEM_PRI           -10
#define SYSTEM_CAN_HANDLE(x) ((x)==SYSTEM_VIACCESS)

//#define DEBUG_LOG

#ifdef DEBUG_LOG
#define dl(x) { (x); }
#else
#define dl(x) ; 
#endif

// -- cPlainKeyVia -------------------------------------------------------------

#define VIA1_KEYLEN 8
#define VIA2_KEYLEN 16
#define VIATPS_KEYLEN 16

class cPlainKeyVia : public cPlainKeyStd {
protected:
  virtual cString PrintKeyNr(void);
public:
  cPlainKeyVia(bool Super);
  virtual bool Parse(const char *line);
  };

static cPlainKeyTypeReg<cPlainKeyVia,'V'> KeyReg;

cPlainKeyVia::cPlainKeyVia(bool Super)
:cPlainKeyStd(Super)
{}

bool cPlainKeyVia::Parse(const char *line)
{
  unsigned char sid[3], skeynr, skey[VIA2_KEYLEN];
#ifdef DEBUG
  const char *sline=line;
#endif
  int len;
  if(GetChar(line,&type,1) && (len=GetHex(line,sid,3,false))) {
     type=toupper(type); id=Bin2Int(sid,len);
     line=skipspace(line);
     bool ok=false;
     if(!strncasecmp(line,"TPS",3)) {
       line+=3;
       keynr=MBC3('T','P','S');
       ok=(len=GetHex(line,skey,VIATPS_KEYLEN));
       }
     else if(GetHex(line,&skeynr,1)) {
       keynr=skeynr;
       ok=((len=GetHex(line,skey,VIA2_KEYLEN,false)) && (len==VIA1_KEYLEN || len==VIA2_KEYLEN));
       }
     if(ok) {
       SetKeyFromBin(skey,len);
       return true;
       }
    }
  d(printf("plainkeyvia: bad key format '%s'\n",sline))
  return false;
}

cString cPlainKeyVia::PrintKeyNr(void)
{
  return cString::sprintf((keynr==MBC3('T','P','S'))?"TPS":"%02X",keynr);
}

// -- cViaccessCardInfo --------------------------------------------------------

class cViaccessCardInfo : public cSimpleItem {
public:
  unsigned char ID[3], UA[5], SA[4], KEYNO[1], KEY[8];
  //
  bool Parse(const char *line);
  bool Save(FILE *f) { return true; }
  bool IsUpdated(void) { return false; }
  void Updated(void) {}
  bool Cmp(cViaccessCardInfo *ci) { return false; }
  };

bool cViaccessCardInfo::Parse(const char *line)
{
  return GetHex(line,ID,sizeof(ID)) && 
         GetHex(line,UA,sizeof(UA)) &&
         GetHex(line,SA,sizeof(SA)) &&
         GetHex(line,KEYNO,sizeof(KEYNO)) && 
         GetHex(line,KEY,sizeof(KEY));
}

// -- cViaccessCardInfos -------------------------------------------------------

class cViaccessCardInfos : public cCardInfos<cViaccessCardInfo> {
public:
  cViaccessCardInfos(void):cCardInfos<cViaccessCardInfo>(SYSTEM_NAME) {}
  };

static cViaccessCardInfos Vcards;

// -- cViaccess ----------------------------------------------------------------

class cViaccess : protected cDes {
private:
  unsigned char v2key[8];
  bool v2mode;
  //
  int HashNanos(const unsigned char *data, int len);
  void Via2Mod(const unsigned char *key2, unsigned char *data);
protected:
  unsigned char hbuff[8], hkey[8];
  int pH;
  //
  void SetV2Mode(const unsigned char *key2);
  void SetHashKey(const unsigned char *key);
  void HashByte(unsigned char c);
  void HashClear(void);
  void Hash(void);
  //
  void Decode(unsigned char *data, const unsigned char *key);
  bool Decrypt(const unsigned char *work_key, const unsigned char *data, int len, unsigned char *des_data1, unsigned char *des_data2);
  //
  virtual unsigned int Mod(unsigned int R, unsigned int key7) const;
public:
  cViaccess(void);
  };

cViaccess::cViaccess(void)
:cDes()
{
  v2mode=false;
}

/* viaccess DES modification */

unsigned int cViaccess::Mod(unsigned int R, unsigned int key7) const
{
  if(key7!=0) {
    const unsigned int key5=(R>>24)&0xff;
    unsigned int al=key7*key5 + key7 + key5;
    al=(al&0xff)-((al>>8)&0xff);
    if(al&0x100) al++;
    R=(R&0x00ffffffL) + (al<<24);
    }
  return R;
}

/* viaccess2 modification. Extracted from "Russian wafer" card.
   A lot of thanks to it's author :) */

void cViaccess::Via2Mod(const unsigned char *key2, unsigned char *data)
{
  int kb, db;
  for(db=7; db>=0; db--) {
    for(kb=7; kb>3; kb--) {
      int a0=kb^db;
      int pos=7;
      if(a0&4) { a0^=7; pos^=7; }
      a0=(a0^(kb&3)) + (kb&3);
      if(!(a0&4)) data[db]^=(key2[kb] ^ ((data[kb^pos]*key2[kb^4]) & 0xFF));
      }
    }
  for(db=0; db<8; db++) {
    for(kb=0; kb<4; kb++) {
      int a0=kb^db;
      int pos=7;
      if(a0&4) { a0^=7; pos^=7; }
      a0=(a0^(kb&3)) + (kb&3);
      if(!(a0&4)) data[db]^=(key2[kb] ^ ((data[kb^pos]*key2[kb^4]) & 0xFF));
      }
    }
}

void cViaccess::Decode(unsigned char *data, const unsigned char *key)
{
  if(v2mode) Via2Mod(v2key,data);
  Des(data,key,VIA_DES);
  if(v2mode) Via2Mod(v2key,data);
}

void cViaccess::SetV2Mode(const unsigned char *key2)
{
  if(key2) {
    memcpy(v2key,key2,sizeof(v2key));
    v2mode=true;
    }
  else v2mode=false;
}


void cViaccess::SetHashKey(const unsigned char *key)
{
  memcpy(hkey,key,sizeof(hkey));
}

void cViaccess::HashByte(unsigned char c)
{
  hbuff[pH++]^=c;
  if(pH==8) { pH=0; Hash(); }
}

void cViaccess::HashClear(void)
{
  memset(hbuff,0,sizeof(hbuff));
  pH=0;
}

void cViaccess::Hash(void)
{
  if(v2mode) Via2Mod(v2key,hbuff);
  Des(hbuff,hkey,VIA_DES_HASH);
  if(v2mode) Via2Mod(v2key,hbuff);
}

int cViaccess::HashNanos(const unsigned char *data, int len)
{
  int i=0;
  pH=0;
  if(data[0]==0x9f) {
    HashByte(data[i++]);
    HashByte(data[i++]);
    for(int j=0; j<data[1]; j++) HashByte(data[i++]);
    while(pH!=0) HashByte(0);
    }
  for(; i<len; i++) HashByte(data[i]);
  return i;
}

bool cViaccess::Decrypt(const unsigned char *work_key, const unsigned char *data, int len, unsigned char *des_data1, unsigned char *des_data2)
{
  int pos=0, encStart=0;
  unsigned char signatur[8];
  while(pos<len) {
    switch(data[pos]) {
      case 0xea:                 // encrypted bytes
        encStart = pos + 2;
        memcpy(des_data1,&data[pos+2],8);
        memcpy(des_data2,&data[pos+2+8],8);
        break;
      case 0xf0:                 // signature
        memcpy(signatur,&data[pos+2],8);
        break;
      }
    pos += data[pos+1]+2;
    }
  HashClear();
  SetHashKey(work_key);
  // key preparation
  unsigned char prepared_key[8];
  if(work_key[7]==0) {
    // 8th key-byte = 0 then like Eurocrypt-M but with viaccess mods
    HashNanos(data,encStart+16);
    memcpy(prepared_key,work_key,sizeof(prepared_key));
    }
  else { // key8 not zero
    // rotate the key 2x left
    prepared_key[0]=work_key[2];
    prepared_key[1]=work_key[3];
    prepared_key[2]=work_key[4];
    prepared_key[3]=work_key[5];
    prepared_key[4]=work_key[6];
    prepared_key[5]=work_key[0];
    prepared_key[6]=work_key[1];
    prepared_key[7]=work_key[7];
    // test if key8 odd
    if(work_key[7]&1) {
      HashNanos(data,encStart);
      // test if low nibble zero
      unsigned char k = ((work_key[7] & 0xf0) == 0) ? 0x5a : 0xa5;
      for(int i=0; i<8; i++) {
        unsigned char tmp=des_data1[i];
        des_data1[i]=(k & hbuff[pH]) ^ tmp;
        HashByte(tmp);
        }
      for(int i=0; i<8; i++) {
        unsigned char tmp=des_data2[i];
        des_data2[i]=(k & hbuff[pH]) ^ tmp;
        HashByte(tmp);
        }
      }
    else {
      HashNanos(data,encStart+16);
      }
    }
  Decode(des_data1,prepared_key);
  Decode(des_data2,prepared_key);
  Hash();
  return (memcmp(signatur,hbuff,8)==0);
}

// -- cTPS ---------------------------------------------------------------------

class cTPS : private cAES {
private:
  bool keySet;
  static const unsigned char T1[], T2[];
protected:
  void SetV2Key(const unsigned char *key);
  int Decrypt(unsigned char *data, int len);
public:
  cTPS(void);
  };

const unsigned char cTPS::T1[] = {
  0x4F,0xB4,0xFC,0x9B,0x4A,0x7F,0x44,0xFB,0x05,0xFF,0xBD,0xBB,0x16,0x2D,0x6C,0xC8,
  0xD8,0x96,0xF9,0xFE,0x3F,0xFF,0x36,0x24,0xB6,0xBF,0x49,0xC9,0x2D,0x36,0x5E,0xD0,
  0x1F,0x09,0x7E,0xA9,0x7F,0xFF,0x64,0xB6,0x5B,0x7E,0xF8,0xFC,0x6E,0x3F,0x7F,0xBF,
  0xDD,0x36,0x12,0xE9,0x05,0xFE,0xB4,0x6C,0x6F,0xFE,0x7E,0xC8,0x25,0x90,0x6D,0x90
  };

const unsigned char cTPS::T2[] = {
  0x7E,0x6D,0x7E,0x12,0x76,0xFD,0x2F,0xFE,0x6D,0xFE,0xDA,0x3F,0xDA,0x6D,0xBD,0x97,
  0xD0,0x6D,0xD8,0x9F,0x69,0xFD,0xB6,0x37,0xFE,0x7F,0x36,0x92,0xBD,0x52,0x16,0xDF,
  0xFC,0x96,0xFF,0x92,0xFD,0x6D,0x7F,0xB5,0xFB,0x4C,0xB6,0xB7,0x7E,0xD9,0xFE,0x9B,
  0xFD,0xF4,0x6D,0x9B,0xB9,0x36,0xBF,0x7F,0xD2,0x2D,0xDF,0xB7,0xD9,0xFE,0x69,0xBF
  };

cTPS::cTPS(void)
{
  keySet=false;
}

void cTPS::SetV2Key(const unsigned char *key)
{
  cAES::SetKey(key); keySet=true;
}

int cTPS::Decrypt(unsigned char *data, int len)
{
  if(data[0]!=0xD2 || data[1]!=0x01 || data[2]!=0x01) return 0;
  data+=3; len-=3;
  if(data[0]==0x40) { // TPS crypt v2 (AES)
    if(!keySet) return -1;
    data[0]=0x90;
    for(int i=0; i<len; i+=data[i+1]+2)
      if(data[i]==0xEA) {
        cAES::Decrypt(&data[i+2],16);
        break;
        }
    return 2;
    }
  else {              // TPS crypt v1 (XOR)
    const unsigned char * const table=(data[0]==0xDF) ? T1 : T2;
    for(int i=len-1 ; i>=0 ; i--) data[i]^=table[i&63];
    return 1;
    }
}

// -- cSystemViaccess ----------------------------------------------------------

struct ViaccessSOIDData {
  unsigned char ident[3];
  unsigned char keyno;
  };

#define MAX_NEW_KEYS 5

class cSystemViaccess : public cSystem, private cViaccess, private cTPS {
private:
  struct ViaccessSOIDData sd;
  unsigned char emmBuff[256];
  int emmLen;
  bool haveSOID;
public:
  cSystemViaccess(void);
  virtual bool ProcessECM(const cEcmInfo *ecm, unsigned char *data);
  virtual void ProcessEMM(int pid, int caid, unsigned char *buffer);
  virtual void ParseCADescriptor(cSimpleList<cEcmInfo> *ecms, unsigned short sysId, const unsigned char *data, int len);
  };

cSystemViaccess::cSystemViaccess(void)
:cSystem(SYSTEM_NAME,SYSTEM_PRI)
{
  haveSOID=false; emmLen=0;
  hasLogger=true;
}

void cSystemViaccess::ParseCADescriptor(cSimpleList<cEcmInfo> *ecms, unsigned short sysId, const unsigned char *data, int len)
{
  const int pid=WORD(data,2,0x1FFF);
  if(pid>=0xAA && pid<=0xCF) {
    d(printf("system-viaccess: dropped \"fake\" ecm pid 0x%04x\n",pid))
    return;
    }
  cSystem::ParseCADescriptor(ecms,sysId,data,len);
}

bool cSystemViaccess::ProcessECM(const cEcmInfo *ecm, unsigned char *data)
{
  int len=SCT_LEN(data);
  unsigned char tmp[len];
  cPlainKey *tpsPk=0;
  cKeySnoop *ks=0;
  while(1) {
    memcpy(tmp,data,len);
    unsigned char *nanos=(unsigned char *)cParseViaccess::NanoStart(tmp);
    int nlen=len-(nanos-tmp);

    int tps=cTPS::Decrypt(nanos,nlen);
    if(tps>0) { nanos+=3; nlen-=3; }

    if(tps>=0 && cParseViaccess::CheckNano90FromNano(nanos)) {
      int keynr=cParseViaccess::KeyNrFromNano(nanos);
      if(!ks) ks=new cKeySnoop(this,'V',ecm->provId,keynr);
      cPlainKey *pk=0;
      while((pk=keys.FindKey('V',ecm->provId,keynr,pk))) {
        unsigned char key[16];
        pk->Get(key);
        SetV2Mode(pk->Size()==VIA2_KEYLEN ? &key[VIA1_KEYLEN] : 0);
        if(cViaccess::Decrypt(key,&nanos[5],nlen-5,&cw[0],&cw[8])) {
          ks->OK(pk); delete ks;
          return true;
          }
        }
      }

    if(tps==1 || tps==0) {
      delete ks;
      return false;
      }

    tpsPk=keys.FindKey('V',ecm->provId,MBC3('T','P','S'),16,tpsPk);
    if(!tpsPk) {
      d(printf("system-viaccess: missing/invalid %.4x TPS or OP key\n",ecm->provId))
      delete ks;
      return false;
      }
    unsigned char key[16];
    tpsPk->Get(key);
    cTPS::SetV2Key(key);
    }
}

void cSystemViaccess::ProcessEMM(int pid, int caid, unsigned char *buffer)
{
  unsigned char ua[5], sa[3], signature[8], addressdata[34];
  unsigned char *scan=0;
  unsigned int updtype=0, scanlen=0;

  bool haveData=false;
  const unsigned int msgLen=buffer[2];
  switch(buffer[0]) {
    case 0x8C: // shared update message
    case 0x8D:
      //dl(printf("logger-viaccess %d: got SOID data\n",CardNum()))
      if(msgLen<=sizeof(emmBuff)) {
        scan=&buffer[3]; // Skip SA
        unsigned int cnt=0;
        while(cnt<msgLen) {
          if(scan[cnt]==0x90 && scan[cnt+1]==0x03) { // SOID data
            sd.ident[0]=scan[cnt+2];
            sd.ident[1]=scan[cnt+3];
            sd.ident[2]=scan[cnt+4]&0xF0;
            sd.keyno=scan[cnt+4]&0x0F;
            haveSOID=true;
            break;
            }
          cnt+=scan[cnt+1]+2;
          }
        memcpy(emmBuff,&buffer[3],msgLen);
        emmLen=msgLen;
        }
      else d(printf("logger-viaccess %d: static EMM buffer overflow: msgLen %d > size %d\n",CardNum(),msgLen,sizeof(emmBuff)))
      return;

    case 0x8E: // shared update data/message
      {

      static const unsigned char addr_nano[] = { 0x9E,0x20 };

      memcpy(sa,&buffer[3],sizeof(sa));
      if((buffer[6]&2)==0) { // constant format, needs section 0x8c or 0x8d
        if(emmLen>0) {       // for messages
          memcpy(addressdata, addr_nano, sizeof(addr_nano));
          memcpy(&addressdata[2],&buffer[7], sizeof(addressdata)-2);
          memcpy(signature, &buffer[msgLen-5], sizeof(signature));
          scan=emmBuff; scanlen=emmLen;
          }
        else {
          //d(printf("logger-viaccess %d: section 0x8e with empty buffer\n",CardNum()))
          return;
          }
        }
      else { // variable format with nanos
        scan=&buffer[7]; // Skip SA
        unsigned int cnt=0;
        while(cnt<msgLen) {
          if(scan[cnt]==0xF0 && scan[cnt+1]==0x08) { // this is the last one in line
            memcpy(signature, &scan[cnt+2], sizeof(signature));
            break;
            }
          if(scan[cnt]==addr_nano[0] && scan[cnt+1]==addr_nano[1]) {
            memcpy(addressdata, &scan[cnt], sizeof(addressdata));             
            }
          // no need to copy the data to emmBuff, we are scanning from
          // buffer anyways
          cnt+=scan[cnt+1]+2;
          }
        scanlen=cnt;
        }
      haveData=true; updtype=1;
      break;
      }

    case 0x88:
      {
      if(msgLen>=0xB4) {
        //d(printf("logger-viaccess %d: got broken section data, ignoring it\n", cardNum))
        return;
        }
      memcpy(ua,&buffer[3],sizeof(ua));
      scan=&buffer[8]; // Skip UA
      unsigned int cnt=0;
      while(cnt<msgLen) {
        if(scan[cnt]==0xF0 && scan[cnt+1]==0x08) { 
          memcpy(signature, &scan[cnt+2], sizeof(signature));
          break;
          }
        if(scan[cnt]==0x90 && scan[cnt+1]==0x03) { // SOID data
          sd.ident[0]=scan[cnt+2];
          sd.ident[1]=scan[cnt+3];
          sd.ident[2]=scan[cnt+4]&0xF0;
          sd.keyno=scan[cnt+4]&0x0F;
          haveSOID=true;
          }
        // no need to copy the data to emmBuff, we are scanning from
        // buffer anyways
        cnt+=scan[cnt+1]+2;
        }
      scanlen=cnt; haveData=true; updtype=2;
      break;
      }

    default:
      return;
    }

  // we always want to have complete data
  // (but in theory we can't get until here without data)
  if(!haveData || !haveSOID) return;
  
#if defined(DEBUG_LOG) | defined(DEBUG_LOG_FILE)
  char str[20], str2[20];
  static char *ptext[] = { "TO ALL","SHARED","UNIQUE" };
  const char *addr = (updtype==1) ? HexStr(str,sa,sizeof(sa)) : HexStr(str,ua,sizeof(ua));
#endif
  //dl(printf("%02x%02x00 %02x %s %s\r",ident[0],ident[1],keyno,HexStr(str,ua,sizeof(ua)),ptext[updtype]))
  //dl(fflush(stdout))

  // we are not allowed to modify scan & scanlen inside the following
  // loop or we won't be able to proceed multiple card with the same ident
  cViaccessCardInfo *mkey=Vcards.First();
  while(mkey) {
    if(   !memcmp(mkey->ID,sd.ident,sizeof(sd.ident))
       && mkey->KEYNO[0]==sd.keyno
       && (   (updtype==1 && !memcmp(mkey->SA,sa,sizeof(sa)))
           || (updtype==2 && !memcmp(mkey->UA,ua,sizeof(ua)))
          )
       ) {
      HashClear();
      SetHashKey(mkey->KEY);
      switch(updtype) {
        case 1: // shared/normal update
          memcpy(hbuff+5,sa,3);
          Hash();
          for(unsigned int i=0; i<sizeof(addressdata); i++)
            HashByte(addressdata[i]);
          Hash();
          pH=0;
          break;
        case 2: // unique update
          memcpy(hbuff+3,ua,sizeof(ua));
          Hash();
          break;
        }
        
      unsigned char newKey[MAX_NEW_KEYS][8];
      int numKeys=0, updPrv[MAX_NEW_KEYS]={}, updKey[MAX_NEW_KEYS]={};

      unsigned int cnt=0;
      while(cnt<scanlen && numKeys<MAX_NEW_KEYS) {
        const unsigned int parm=scan[cnt++];
        unsigned int plen=scan[cnt++];

        switch(parm) {
          case 0x90:
            // we shouldn't process nano 0x90 (SOID) in any way since we parsed it before
            cnt+=plen;
            break;

          case 0x9E:
            // we shouldn't process nano 0x9e in any way since we added it to
            // the hash earlier.  We don't prevent it from being in emmBuff
            // anymore, so we just skip it and get the next nano. This is most
            // important for constant shared updates else the hash will be screwed.
            cnt+=plen;
            break;

          case 0xA1: // keyupdate
            updPrv[numKeys]=(scan[cnt]<<16)+(scan[cnt+1]<<8)+(scan[cnt+2]&0xF0);
            updKey[numKeys]=scan[cnt+2]&0x0F;
            // fall through to "default"

          default:
            HashByte(parm); HashByte(plen);
            while(plen--) HashByte(scan[cnt++]);
            break;

          case 0xEF: // crypted key(s)
            HashByte(parm); HashByte(plen);
            if(plen==sizeof(newKey[0])) {
              const unsigned char k7=mkey->KEY[7];
              for(unsigned int kc=0 ; kc<sizeof(newKey[0]) ; kc++) {
                const unsigned char b=scan[cnt++];
                if(k7&1) newKey[numKeys][kc]=b^(hbuff[pH]&(k7<0x10 ? 0x5a : 0xa5));
                else     newKey[numKeys][kc]=b;
                HashByte(b);
                }
              numKeys++;
              }
            else {
              d(printf("logger-viaccess %d: key length mismatch %d!=%d\n",CardNum(),plen,sizeof(newKey[0])))
              return;
              }
            break;
          }
        }
      Hash();

      if(numKeys) {
        if(!memcmp(signature,hbuff,sizeof(signature))) {
          unsigned char key[8];
          memcpy(key,mkey->KEY,sizeof(key));
          if(key[7]) { // Rotate key
            const unsigned char t1=key[0], t2=key[1];
            key[0]=key[2]; key[1]=key[3]; key[2]=key[4]; key[3]=key[5]; key[4]=key[6];
            key[5]=t1; key[6]=t2;
            }

          while(numKeys--) {
            dl(printf("%02x%02x %02x %s %s - KEY %06x.%02x",sd.ident[0],sd.ident[1],
              sd.keyno,addr,ptext[updtype],updPrv[numKeys],updKey[numKeys]))
            Decode(newKey[numKeys],key);
            dl(printf(" -> %s\n",HexStr(str2,newKey[numKeys],sizeof(newKey[numKeys]))))
#ifdef DEBUG_LOG_FILE
            FILE *f=LogStart();
            if (f) {
              fprintf(f,"%02x%02x %02x %s %s - KEY %06x.%02x -> %s\n",
                        sd.ident[0],sd.ident[1],sd.keyno,addr,ptext[updtype],
                        updPrv[numKeys],updKey[numKeys],HexStr(str2,newKey[numKeys],sizeof(newKey[numKeys])));
              LogEnd(f);
              }
#endif
            cSoftCAM::FoundKey();
            if(keys.NewKey('V',updPrv[numKeys],updKey[numKeys],newKey[numKeys],8))
              cSoftCAM::NewKey();
            }
          cSoftCAM::Save();
          }
        else {
          dl(printf("%02x%02x %02x %s %s - FAIL\n",
                     sd.ident[0],sd.ident[1],sd.keyno,addr,ptext[updtype]))

#ifdef DEBUG_LOG_FILE
          FILE *f=LogStart();
          if (f) {
            fprintf(f,"%02x%02x %02x %s %s - FAIL\n",
                      sd.ident[0],sd.ident[1],sd.keyno,addr,ptext[updtype]);
            LogEnd(f);
            }
#endif
          }
        }
      }
    mkey=Vcards.Next(mkey);
    }
}

// -- cSystemLinkViaccess ------------------------------------------------------

class cSystemLinkViaccess : public cSystemLink {
public:
  cSystemLinkViaccess(void);
  virtual bool CanHandle(unsigned short SysId);
  virtual cSystem *Create(void) { return new cSystemViaccess; }
  virtual bool Init(const char *cfgdir);
  };

static cSystemLinkViaccess staticInit;

cSystemLinkViaccess::cSystemLinkViaccess(void)
:cSystemLink(SYSTEM_NAME,SYSTEM_PRI)
{
  Feature.NeedsKeyFile();
}

bool cSystemLinkViaccess::CanHandle(unsigned short SysId)
{
  SysId&=SYSTEM_MASK;
  return SYSTEM_CAN_HANDLE(SysId);
}

bool cSystemLinkViaccess::Init(const char *cfgdir)
{
  Vcards.Load(cfgdir,SYSTEM_NAME,"Viaccess.KID");
  return true;
}
